@@ -1,4 +1,4 @@
 from ._scholarly import _Scholarly
-from ._proxy_generator import ProxyGenerator
 from .data_types import Author, Publication
+from ._proxy_generator import ProxyGenerator, DOSException, MaxTriesExceededException
 scholarly = _Scholarly()
